# Copyright 2017 The Bazel Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pyformat: disable
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# pylint: skip-file
# source: src/main/protobuf/command_line.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import option_filters_pb2 as src_dot_main_dot_protobuf_dot_option__filters__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='src/main/protobuf/command_line.proto',
  package='command_line',
  syntax='proto3',
  serialized_options=b'\n+com.google.devtools.build.lib.runtime.proto',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n$src/main/protobuf/command_line.proto\x12\x0c\x63ommand_line\x1a&src/main/protobuf/option_filters.proto\"]\n\x0b\x43ommandLine\x12\x1a\n\x12\x63ommand_line_label\x18\x01 \x01(\t\x12\x32\n\x08sections\x18\x02 \x03(\x0b\x32 .command_line.CommandLineSection\"\x9b\x01\n\x12\x43ommandLineSection\x12\x15\n\rsection_label\x18\x01 \x01(\t\x12-\n\nchunk_list\x18\x02 \x01(\x0b\x32\x17.command_line.ChunkListH\x00\x12/\n\x0boption_list\x18\x03 \x01(\x0b\x32\x18.command_line.OptionListH\x00\x42\x0e\n\x0csection_type\"\x1a\n\tChunkList\x12\r\n\x05\x63hunk\x18\x01 \x03(\t\"2\n\nOptionList\x12$\n\x06option\x18\x01 \x03(\x0b\x32\x14.command_line.Option\"\xac\x01\n\x06Option\x12\x15\n\rcombined_form\x18\x01 \x01(\t\x12\x13\n\x0boption_name\x18\x02 \x01(\t\x12\x14\n\x0coption_value\x18\x03 \x01(\t\x12-\n\x0b\x65\x66\x66\x65\x63t_tags\x18\x04 \x03(\x0e\x32\x18.options.OptionEffectTag\x12\x31\n\rmetadata_tags\x18\x05 \x03(\x0e\x32\x1a.options.OptionMetadataTagB-\n+com.google.devtools.build.lib.runtime.protob\x06proto3'
  ,
  dependencies=[src_dot_main_dot_protobuf_dot_option__filters__pb2.DESCRIPTOR,])




_COMMANDLINE = _descriptor.Descriptor(
  name='CommandLine',
  full_name='command_line.CommandLine',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='command_line_label', full_name='command_line.CommandLine.command_line_label', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sections', full_name='command_line.CommandLine.sections', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=94,
  serialized_end=187,
)


_COMMANDLINESECTION = _descriptor.Descriptor(
  name='CommandLineSection',
  full_name='command_line.CommandLineSection',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='section_label', full_name='command_line.CommandLineSection.section_label', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='chunk_list', full_name='command_line.CommandLineSection.chunk_list', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='option_list', full_name='command_line.CommandLineSection.option_list', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='section_type', full_name='command_line.CommandLineSection.section_type',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=190,
  serialized_end=345,
)


_CHUNKLIST = _descriptor.Descriptor(
  name='ChunkList',
  full_name='command_line.ChunkList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='chunk', full_name='command_line.ChunkList.chunk', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=347,
  serialized_end=373,
)


_OPTIONLIST = _descriptor.Descriptor(
  name='OptionList',
  full_name='command_line.OptionList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='option', full_name='command_line.OptionList.option', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=375,
  serialized_end=425,
)


_OPTION = _descriptor.Descriptor(
  name='Option',
  full_name='command_line.Option',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='combined_form', full_name='command_line.Option.combined_form', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='option_name', full_name='command_line.Option.option_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='option_value', full_name='command_line.Option.option_value', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='effect_tags', full_name='command_line.Option.effect_tags', index=3,
      number=4, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metadata_tags', full_name='command_line.Option.metadata_tags', index=4,
      number=5, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=428,
  serialized_end=600,
)

_COMMANDLINE.fields_by_name['sections'].message_type = _COMMANDLINESECTION
_COMMANDLINESECTION.fields_by_name['chunk_list'].message_type = _CHUNKLIST
_COMMANDLINESECTION.fields_by_name['option_list'].message_type = _OPTIONLIST
_COMMANDLINESECTION.oneofs_by_name['section_type'].fields.append(
  _COMMANDLINESECTION.fields_by_name['chunk_list'])
_COMMANDLINESECTION.fields_by_name['chunk_list'].containing_oneof = _COMMANDLINESECTION.oneofs_by_name['section_type']
_COMMANDLINESECTION.oneofs_by_name['section_type'].fields.append(
  _COMMANDLINESECTION.fields_by_name['option_list'])
_COMMANDLINESECTION.fields_by_name['option_list'].containing_oneof = _COMMANDLINESECTION.oneofs_by_name['section_type']
_OPTIONLIST.fields_by_name['option'].message_type = _OPTION
_OPTION.fields_by_name['effect_tags'].enum_type = src_dot_main_dot_protobuf_dot_option__filters__pb2._OPTIONEFFECTTAG
_OPTION.fields_by_name['metadata_tags'].enum_type = src_dot_main_dot_protobuf_dot_option__filters__pb2._OPTIONMETADATATAG
DESCRIPTOR.message_types_by_name['CommandLine'] = _COMMANDLINE
DESCRIPTOR.message_types_by_name['CommandLineSection'] = _COMMANDLINESECTION
DESCRIPTOR.message_types_by_name['ChunkList'] = _CHUNKLIST
DESCRIPTOR.message_types_by_name['OptionList'] = _OPTIONLIST
DESCRIPTOR.message_types_by_name['Option'] = _OPTION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CommandLine = _reflection.GeneratedProtocolMessageType('CommandLine', (_message.Message,), {
  'DESCRIPTOR' : _COMMANDLINE,
  '__module__' : 'src.main.protobuf.command_line_pb2'
  # @@protoc_insertion_point(class_scope:command_line.CommandLine)
  })
_sym_db.RegisterMessage(CommandLine)

CommandLineSection = _reflection.GeneratedProtocolMessageType('CommandLineSection', (_message.Message,), {
  'DESCRIPTOR' : _COMMANDLINESECTION,
  '__module__' : 'src.main.protobuf.command_line_pb2'
  # @@protoc_insertion_point(class_scope:command_line.CommandLineSection)
  })
_sym_db.RegisterMessage(CommandLineSection)

ChunkList = _reflection.GeneratedProtocolMessageType('ChunkList', (_message.Message,), {
  'DESCRIPTOR' : _CHUNKLIST,
  '__module__' : 'src.main.protobuf.command_line_pb2'
  # @@protoc_insertion_point(class_scope:command_line.ChunkList)
  })
_sym_db.RegisterMessage(ChunkList)

OptionList = _reflection.GeneratedProtocolMessageType('OptionList', (_message.Message,), {
  'DESCRIPTOR' : _OPTIONLIST,
  '__module__' : 'src.main.protobuf.command_line_pb2'
  # @@protoc_insertion_point(class_scope:command_line.OptionList)
  })
_sym_db.RegisterMessage(OptionList)

Option = _reflection.GeneratedProtocolMessageType('Option', (_message.Message,), {
  'DESCRIPTOR' : _OPTION,
  '__module__' : 'src.main.protobuf.command_line_pb2'
  # @@protoc_insertion_point(class_scope:command_line.Option)
  })
_sym_db.RegisterMessage(Option)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Local backend executors."""

from typing import Dict, Optional

import attr
from xmanager import xm
from xmanager.docker import docker_adapter

GOOGLE_KUBERNETES_ENGINE_CLOUD_PROVIDER = 'GOOGLE_KUBERNETES_ENGINE'


@attr.s(auto_attribs=True)
class LocalSpec(xm.ExecutorSpec):
  """Current machine executor's specification."""


@attr.s(auto_attribs=True)
class DockerOptions:
  """Options of the container to be run.

  Attributes:
      ports: In the simplest form -- a dictionary from `int` to `int`, where the
        keys represent the ports inside the container and the values represent
        the ports of the host to bind. See the specification at
        https://docker-py.readthedocs.io/en/stable/containers.html.
      volumes: A dictionary from `str` to `str`, where the keys represent paths
        inside on the host to mount and the values represent paths in the
        container.
  """
  ports: Optional[docker_adapter.Ports] = None
  volumes: Optional[Dict[str, str]] = None


@attr.s(auto_attribs=True)
class Local(xm.Executor):
  """Current machine executor.

  Attributes:
    docker_options: Options applied if the job is a container-based executable.
    experimental_stream_output: Whether to pipe the job's stdout and stderr to
      the terminal. Might be removed once we decide on the logging design.
  """

  docker_options: Optional[DockerOptions] = None
  experimental_stream_output: bool = False

  Spec = LocalSpec  # pylint: disable=invalid-name


@attr.s(auto_attribs=True)
class TpuCapability:
  """TPU capability configures the TPU software requested by an executor."""

  # Read about TPU versions:
  # https://cloud.google.com/tpu/docs/version-switching
  tpu_runtime_version: str


@attr.s(auto_attribs=True)
class TensorboardCapability:
  """Tensorboard capability integrates a CAIP Job with Vertex Tensorboard."""

  # The name of the tensorboard to use.
  name: str
  # The "gs://$GCS_BUCKET/dir_name" to save output.
  # Tensorboard will read the logs from $BASE_OUTPUT_DIRECTORY/logs/
  # If None, then the root of the default bucket will be used.
  base_output_directory: Optional[str] = None


@attr.s(auto_attribs=True)
class CaipSpec(xm.ExecutorSpec):
  """Caip spec describes the Google Cloud Platform (GCP) location."""

  # An image registry name tag to push.
  # The image tag should be in the form 'myregistryhost/name:tag'
  push_image_tag: Optional[str] = None


@attr.s(auto_attribs=True)
class Caip(xm.Executor):
  """Caip Executor describes the runtime environment of GCP."""

  requirements: xm.JobRequirements = attr.Factory(xm.JobRequirements)
  tpu_capability: Optional[TpuCapability] = None
  tensorboard: Optional[TensorboardCapability] = None

  Spec = CaipSpec  # pylint: disable=invalid-name


@attr.s(auto_attribs=True)
class KubernetesSpec(xm.ExecutorSpec):
  """K8s spec describes the K8s location."""

  # An image registry name tag to push.
  # The image tag should be in the form 'myregistryhost/name:tag'
  push_image_tag: Optional[str] = None


@attr.s(auto_attribs=True)
class Kubernetes(xm.Executor):
  """K8s Executor describes the runtime environment of Kubernetes."""

  requirements: xm.JobRequirements = attr.Factory(xm.JobRequirements)
  cloud_provider: str = GOOGLE_KUBERNETES_ENGINE_CLOUD_PROVIDER
  tpu_capability: Optional[TpuCapability] = None

  Spec = KubernetesSpec  # pylint: disable=invalid-name



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Data classes for job-related abstractions."""

import abc
import itertools
from typing import Any, Awaitable, Callable, Dict, Iterable, List, Mapping, Optional, Sequence, Tuple, TypeVar, Union

import attr

from xmanager.xm import pattern_matching

UserArgs = Union[Mapping, Sequence, 'SequentialArgs']


class SequentialArgs:
  """A mix of positional and keyword arguments.

  Its main purpose is to provide merging capabilities as arguments can be both
  lists and dicts. We cannot just convert dicts to lists as we want to allow
  overriding, for example `['--a'] ∪ {'z': 1, 'y': 2} ∪ ['--b'] ∪ {'z': 3}`
  is `['--a', '--z', '3', '--y', '2', '--b']`.
  """

  @attr.s(auto_attribs=True)
  class _RegularItem:
    value: Any

  @attr.s(auto_attribs=True)
  class _KeywordItem:
    name: str

  def __init__(self):
    self._items: List[Union[SequentialArgs._RegularItem,
                            SequentialArgs._KeywordItem]] = []
    self._kwvalues: Dict[str, Any] = {}

  def _ingest_regular_item(self, value: Any) -> None:
    self._items.append(SequentialArgs._RegularItem(value))

  def _ingest_keyword_item(self, name: str, value: Any) -> None:
    if name not in self._kwvalues:
      self._items.append(SequentialArgs._KeywordItem(name))
    self._kwvalues[name] = value

  def _merge_from(self, args: 'SequentialArgs') -> None:
    """Merges another instance of SequentialArgs into self."""

    def import_regular_item(item: SequentialArgs._RegularItem):
      self._ingest_regular_item(item.value)

    def import_keyword_item(item: SequentialArgs._KeywordItem):
      self._ingest_keyword_item(item.name, args._kwvalues[item.name])  # pylint: disable=protected-access

    matcher = pattern_matching.match(
        import_regular_item,
        import_keyword_item,
    )
    for item in args._items:  # pylint: disable=protected-access
      matcher(item)

  @staticmethod
  def merge(operands: Iterable['SequentialArgs']) -> 'SequentialArgs':
    """Merges several instances into one left-to-right."""
    result = SequentialArgs()
    for operand in operands:
      result._merge_from(operand)  # pylint: disable=protected-access
    return result

  @staticmethod
  def from_collection(collection: Optional[UserArgs]) -> 'SequentialArgs':
    """Populates a new instance from a given collection."""
    result = SequentialArgs()
    if collection is None:
      return result

    def import_sequential_args(args: SequentialArgs) -> None:
      result._merge_from(args)  # pylint: disable=protected-access

    def import_mapping(collection: Mapping[Any, Any]) -> None:
      for key, value in collection.items():
        result._ingest_keyword_item(str(key), value)  # pylint: disable=protected-access

    def import_sequence(collection: Sequence[Any]) -> None:
      for value in collection:
        result._ingest_regular_item(value)  # pylint: disable=protected-access

    matcher = pattern_matching.match(import_sequential_args, import_mapping,
                                     import_sequence)
    matcher(collection)
    return result

  def to_list(self, escaper: Callable[[Any], str]) -> List[str]:
    """Exports items as a list ready to be passed into the command line."""

    def export_regular_item(item: SequentialArgs._RegularItem) -> List[str]:
      return [escaper(item.value)]

    def export_keyword_item(item: SequentialArgs._KeywordItem) -> List[str]:
      value = self._kwvalues[item.name]
      if isinstance(value, bool):
        return [escaper(f"--{'' if value else 'no'}{item.name}")]
      else:
        return [escaper(f'--{item.name}'), escaper(value)]

    matcher = pattern_matching.match(
        export_regular_item,
        export_keyword_item,
    )
    return list(itertools.chain(*[matcher(item) for item in self._items]))

  def to_dict(self, kwargs_only: bool = False) -> Dict[str, Any]:
    """Exports items as a dictionary.

    Args:
      kwargs_only: Whether to skip positional arguments.

    Returns:
      The sought dictionary.
    """
    if kwargs_only:
      return self._kwvalues

    def export_regular_item(
        item: SequentialArgs._RegularItem) -> Tuple[str, Any]:
      return (str(item.value), True)

    def export_keyword_item(
        item: SequentialArgs._KeywordItem) -> Tuple[str, Any]:
      return (item.name, self._kwvalues[item.name])

    matcher = pattern_matching.match(
        export_regular_item,
        export_keyword_item,
    )
    return dict([matcher(item) for item in self._items])

  def __eq__(self, other) -> bool:
    return isinstance(other, SequentialArgs) and all([
        self._items == other._items,
        self._kwvalues == other._kwvalues,
    ])

  def __repr__(self) -> str:
    return f"[{', '.join(self.to_list(repr))}]"


def merge_args(*operands: Union[SequentialArgs, UserArgs]) -> SequentialArgs:
  return SequentialArgs.merge(
      map(
          pattern_matching.match(
              pattern_matching.Case([SequentialArgs], lambda args: args),
              pattern_matching.Case([Any], SequentialArgs.from_collection)),
          operands))


class ExecutableSpec(abc.ABC):
  """Executable specification describes what code / computation to run.

  Use one of the functions declared in xm/packagables.py to create a spec:
   * xm.binary - a prebuilt executable program.
   * xm.bazel_binary - an executable built with Bazel.
   * xm.container - a prebuilt Docker container.
   * xm.bazel_container - a Docker container built with Bazel.
   * xm.python_container - a Docker container running python code.
   * xm.dockerfile_container - a Docker container built with dockerfile.

  An executable spec must be turned into an Executable using
  Experiment.package() in order to be used in a Job.

  WARNING: `ExecutableSpec`s are supposed to be implementation-agnostic. That
  means there should be no backend-specific class inheriting `ExecutableSpec`.
  """

  @property
  @abc.abstractmethod
  def name(self) -> str:
    raise NotImplementedError


@attr.s(auto_attribs=True)
class Executable(abc.ABC):
  """Executable describes the final location of a packaged executable spec.

  An executable depends on the executable specification and the executor
  specification. Experiment's implementation knows how to handle each type of
  executable.

  Attributes:
    name: An automatically populated name for the executable. Used for assigning
      default names to `Job`s.
  """

  name: str


class ExecutorSpec(abc.ABC):
  """Executor spec describes the location of the runtime environment.

  For a list of supported ExecutorSpecs see a list of executors below.
  """


class Executor(abc.ABC):
  """Executor describes the runtime environment of a Job.

  Concrete supported executors are listed in xm_local/executors.py:
    * xm_local.Local
    * xm_local.Caip
    * xm_local.Kubernetes
  """

  @classmethod
  @abc.abstractmethod
  def Spec(cls) -> ExecutorSpec:  # pylint: disable=invalid-name
    raise NotImplementedError


@attr.s(auto_attribs=True)
class Packageable:
  """Packageable describes what to build and its static parameters."""

  executable_spec: ExecutableSpec
  executor_spec: ExecutorSpec
  args: SequentialArgs = attr.ib(
      factory=list, converter=SequentialArgs.from_collection)  # pytype: disable=annotation-type-mismatch
  env_vars: Dict[str, str] = attr.ib(converter=dict, default=attr.Factory(dict))


class Constraint(abc.ABC):
  """Constraint describes the requirements for where a job group can run.

  Some examples of constraints include:

    * same virtual machine;
    * same virtual private Cloud subnetwork;
    * same network fabric;
    * same geographic location.
  """


JobGeneratorType = Callable[..., Awaitable]
JobType = Union['Job', 'JobGroup', JobGeneratorType]


@attr.s(auto_attribs=True)
class Job:
  """Job describes a unit of computation to be run.

  Attributes:
    executable: What to run -- one of `xm.Experiment.package` results.
    executor: Where to run -- one of `xm.Executor` subclasses.
    name: Name of the job. Must be unique within the context (work unit). By
      default it is constructed from the executable. Used for naming related
      entities such as newly created containers.
    args: Command line arguments to pass.
    env_vars: Environment variables to apply.
  """

  executable: Executable
  executor: Executor
  name: Optional[str] = None
  args: SequentialArgs = attr.ib(
      factory=list, converter=SequentialArgs.from_collection)  # pytype: disable=annotation-type-mismatch
  env_vars: Dict[str, str] = attr.Factory(dict)


class JobGroup:
  """JobGroup describes a set of jobs that run under shared constraints.

  Use named arguments to give jobs meaningful names:

  ```
  JobGroup(
      learner=Job(learner_executable, executor),
      actor=Job(actor_executable, executor),
  )
  ```

  JobGroups provide the gang scheduling concept: Jobs inside them would be
  scheduled / descheduled simultaneously. Note that schedulers may not always be
  able to enforce that.

  JobGroups may include more fine grained constraints:

  ```
  JobGroup(
      learner=Job(tpu_learner_executable, executor),
      preprocessor=Job(preprocessor_executable, executor),
      constraints=[xm_impl.SameMachine()],
  )
  ```

  To express sophisticated requirements JobGroups can be nested:

  ```
  JobGroup(
      eval=Job(eval_executable, executor),
      colocated_learner_and_actor=JobGroup(
          learner=Job(tpu_learner_executable, executor),
          actor=Job(actor_executable, executor),
          constraints=[xm_impl.SameMachine()],
      ),
  )
  ```

  Attributes:
    jobs: A mapping of names to jobs.
    constraints: A list of additional scheduling constraints.
  """

  jobs: Dict[str, JobType]
  constraints: List[Constraint]

  def __init__(self,
               *,
               constraints: Optional[Sequence[Constraint]] = None,
               **jobs: JobType) -> None:
    """Builds a JobGroup.

    Args:
      constraints: List of additional scheduling constraints. Keyword only arg.
      **jobs: Jobs / job groups that constitute the group passed as kwargs.
    """
    self.jobs = jobs
    self.constraints = list(constraints) if constraints else []


JobTypeVar = TypeVar('JobTypeVar', Job, JobGroup, JobGeneratorType)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for xmanager.cloud.utils."""

import os
import tempfile
import unittest

from xmanager.cloud import utils

_CLUSTER_SPEC = """{
  "cluster": {
    "workerpool0": ["cmle-training-workerpool0-ab-0:2222"],
    "workerpool1": ["cmle-training-workerpool1-ab-0:2222", "cmle-training-workerpool1-ab-1:2222"],
    "workerpool2": ["cmle-training-workerpool2-ab-0:2222", "cmle-training-workerpool2-ab-1:2222"]
  },
  "environment": "cloud",
  "task": {
    "type": "workerpool1",
    "index": 1,
    "trial": ""
  }
}""".replace('\n', ' ')


class UtilsTest(unittest.TestCase):

  def tearDown(self):
    super(UtilsTest, self).tearDown()
    os.environ['CLUSTER_SPEC'] = ''

  def test_get_master_address_port(self):
    os.environ['CLUSTER_SPEC'] = _CLUSTER_SPEC
    address, port = utils.get_master_address_port()
    self.assertEqual(address, 'cmle-training-workerpool0-ab-0')
    self.assertEqual(port, '2222')

  def test_get_master_address_port_default(self):
    address, port = utils.get_master_address_port()
    self.assertEqual(address, '127.0.0.1')
    self.assertEqual(port, '29500')

  def test_get_world_size_rank(self):
    os.environ['CLUSTER_SPEC'] = _CLUSTER_SPEC
    world_size, rank = utils.get_world_size_rank()
    self.assertEqual(world_size, 5)
    self.assertEqual(rank, 2)

  def test_get_world_size_rank_default(self):
    world_size, rank = utils.get_world_size_rank()
    self.assertEqual(world_size, 1)
    self.assertEqual(rank, 0)

  def test_wrap_and_unwrap_addresses(self):
    arg = '--master=' + utils.get_workerpool_address('workerpool0')
    self.assertEqual(arg, '--master=%objectname(workerpool0)%')
    os.environ['CLUSTER_SPEC'] = _CLUSTER_SPEC
    self.assertEqual(
        utils.map_workerpool_address_args([arg]),
        ['--master=cmle-training-workerpool0-ab-0:2222'])

  def test_create_workerpool_address_env_vars_script(self):
    os.environ['MY_WORKER'] = utils.get_workerpool_address('workerpool0')
    os.environ['CLUSTER_SPEC'] = _CLUSTER_SPEC
    t = tempfile.NamedTemporaryFile()
    utils.create_workerpool_address_env_vars_script(t.name)
    expected = """
#!/bin/bash

export MY_WORKER=cmle-training-workerpool0-ab-0:2222
    """
    with open(t.name) as f:
      self.assertEqual(f.read(), expected.strip())


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Packaging for local executions."""

import os
from typing import Any

from xmanager import xm
from xmanager.bazel import client as bazel_client
from xmanager.cloud import build_image
from xmanager.cloud import docker_lib
from xmanager.docker import docker_adapter
from xmanager.xm import executables
from xmanager.xm import pattern_matching
from xmanager.xm_local import executables as local_executables


def _package_container(packageable: xm.Packageable,
                       container: executables.Container) -> xm.Executable:
  """Packages a container for local execution."""
  instance = docker_adapter.instance()
  image_id = None
  if os.path.exists(container.image_path):
    image_id = instance.load_image(container.image_path)
  elif instance.is_registry_label(container.image_path):
    image_id = instance.pull_image(container.image_path)
  if image_id is not None:
    return local_executables.LoadedContainerImage(
        name=packageable.executable_spec.name,
        image_id=image_id,
        args=packageable.args,
        env_vars=packageable.env_vars,
    )
  else:
    raise ValueError(
        f'{container.image_path} is found neither locally nor remotely')


def _package_binary(packageable: xm.Packageable, binary: executables.Binary):
  if not os.path.exists(binary.path):
    raise ValueError(f'{binary.path} does not exist on this machine')
  return local_executables.LocalBinary(
      name=packageable.executable_spec.name,
      path=binary.path,
      args=packageable.args,
      env_vars=packageable.env_vars,
  )


def _package_dockerfile(packageable: xm.Packageable,
                        dockerfile: executables.Dockerfile):
  image_id = docker_lib.build_docker_image(dockerfile.name, dockerfile.path,
                                           dockerfile.dockerfile)
  return local_executables.LoadedContainerImage(
      name=packageable.executable_spec.name,
      image_id=image_id,
      args=packageable.args,
      env_vars=packageable.env_vars,
  )


def _package_python_container(packageable: xm.Packageable,
                              py_executable: executables.PythonContainer):
  # Use the directory as the image name.
  image_name = os.path.basename(py_executable.path)
  image_id = build_image.build(py_executable, packageable.args,
                               packageable.env_vars, image_name)
  return local_executables.LoadedContainerImage(
      name=packageable.executable_spec.name, image_id=image_id)


def _package_bazel_container(
    packageable: xm.Packageable,
    container: executables.BazelContainer) -> xm.Executable:
  paths = bazel_client.build_single_target(container.label)
  assert len(paths) == 1
  image_id = docker_adapter.instance().load_image(paths[0])
  return local_executables.LoadedContainerImage(
      name=packageable.executable_spec.name,
      image_id=image_id,
      args=packageable.args,
      env_vars=packageable.env_vars)


def _package_bazel_binary(packageable: xm.Packageable,
                          binary: executables.BazelBinary) -> xm.Executable:
  paths = bazel_client.build_single_target(
      label=binary.label,
      tail_args=binary.bazel_args,
  )
  assert len(paths) == 1
  return local_executables.LocalBinary(
      name=packageable.executable_spec.name,
      path=paths[0],
      args=packageable.args,
      env_vars=packageable.env_vars,
  )


def _throw_on_unknown_executable(packageable: xm.Packageable, executable: Any):
  raise TypeError('Unsupported executable specification '
                  f'for local packaging: {executable!r}')


_LOCAL_PACKAGING_ROUTER = pattern_matching.match(
    _package_bazel_binary,
    _package_bazel_container,
    _package_binary,
    _package_container,
    _package_dockerfile,
    _package_python_container,
    _throw_on_unknown_executable,
)


def package_for_local_executor(packageable: xm.Packageable,
                               executable_spec: xm.ExecutableSpec):
  return _LOCAL_PACKAGING_ROUTER(packageable, executable_spec)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for xmanager.cloud.caip."""
import os
import unittest
from unittest import mock

from google import auth
from google.auth import credentials
from google.cloud import aiplatform
from google.cloud import aiplatform_v1 as aip_v1
from google.cloud.aiplatform import utils as aip_utils

from xmanager import xm
from xmanager.cloud import auth as xm_auth
from xmanager.xm_local import executables as local_executables
from xmanager.xm_local import executors as local_executors

from xmanager.cloud import caip  # pylint: disable=g-bad-import-order


class CaipTest(unittest.TestCase):

  @mock.patch.object(xm_auth, 'get_service_account')
  @mock.patch.object(auth, 'default')
  def test_launch(self, mock_creds, mock_sa):
    os.environ['GOOGLE_CLOUD_BUCKET_NAME'] = 'test-bucket'
    creds = credentials.AnonymousCredentials()
    mock_creds.return_value = (creds, 'test-project')
    mock_sa.return_value = 'test-sa'

    client = caip.Client('test-project', 'us-central1')
    job = xm.Job(
        name='test-job',
        executable=local_executables.GoogleContainerRegistryImage(
            name='test-image',
            image_path='image-path',
            args=xm.SequentialArgs.from_collection({'a': 1}),
        ),
        executor=local_executors.Caip(xm.JobRequirements(cpu=1, ram=1, t4=2)),
        args={
            'b': 2,
            'c': 3
        },
    )

    expected_call = {
        'parent':
            'projects/test-project/locations/us-central1',
        'custom_job':
            aip_v1.CustomJob(
                display_name='test-experiment',
                job_spec=aip_v1.CustomJobSpec(
                    worker_pool_specs=[
                        aip_v1.WorkerPoolSpec(
                            machine_spec=aip_v1.MachineSpec(
                                machine_type='n1-highmem-2',
                                accelerator_type='NVIDIA_TESLA_T4',
                                accelerator_count=2,
                            ),
                            replica_count=1,
                            container_spec=aip_v1.ContainerSpec(
                                image_uri='image-path',
                                args=['--a', '1', '--b', '2', '--c', '3'],
                            ))
                    ],
                    service_account='test-sa',
                    base_output_directory=aip_v1.GcsDestination(
                        output_uri_prefix='gs://test-bucket',),
                ),
            ),
    }

    with mock.patch.object(aip_utils.ClientWithOverride, 'WrappedClient') as job_client, \
         mock.patch.object(aiplatform.CustomJob, 'resource_name', new_callable=mock.PropertyMock) as name, \
         mock.patch.object(aiplatform.CustomJob, '_dashboard_uri'):
      name.return_value = 'test-resource-name'
      client.launch('test-experiment', [job])
      job_client.return_value.create_custom_job.assert_called_once_with(
          **expected_call)

  def test_get_machine_spec_default(self):
    job = xm.Job(
        executable=local_executables.GoogleContainerRegistryImage('name', ''),
        executor=local_executors.Caip(),
        args={})
    machine_spec = caip.get_machine_spec(job)
    self.assertDictEqual(machine_spec, {'machine_type': 'n1-standard-4'})

  def test_get_machine_spec_cpu(self):
    job = xm.Job(
        executable=local_executables.GoogleContainerRegistryImage('name', ''),
        executor=local_executors.Caip(
            requirements=xm.JobRequirements(cpu=20, ram=40 * xm.GiB)),
        args={})
    machine_spec = caip.get_machine_spec(job)
    self.assertDictEqual(machine_spec, {'machine_type': 'n1-highcpu-64'})

  def test_get_machine_spec_gpu(self):
    job = xm.Job(
        executable=local_executables.GoogleContainerRegistryImage('name', ''),
        executor=local_executors.Caip(requirements=xm.JobRequirements(p100=2)),
        args={})
    machine_spec = caip.get_machine_spec(job)
    self.assertDictEqual(
        machine_spec, {
            'machine_type': 'n1-standard-4',
            'accelerator_type': caip.aip_v1.AcceleratorType.NVIDIA_TESLA_P100,
            'accelerator_count': 2,
        })

  def test_get_machine_spec_tpu(self):
    job = xm.Job(
        executable=local_executables.GoogleContainerRegistryImage('name', ''),
        executor=local_executors.Caip(
            requirements=xm.JobRequirements(tpu_v3=8)),
        args={})
    machine_spec = caip.get_machine_spec(job)
    # TPU_V2 and TPU_V3 removed in
    # https://github.com/googleapis/python-aiplatform/commit/f3a3d03c8509dc49c24139155a572dacbe954f66
    # When TPU enums are restored, replace
    #   'accelerator_type': 7,
    # with
    #   'accelerator_type': caip.aip_v1.AcceleratorType.TPU_V3,
    self.assertDictEqual(
        machine_spec, {
            'machine_type': 'n1-standard-4',
            'accelerator_type': 7,
            'accelerator_count': 8,
        })

  def test_cpu_ram_to_machine_type_exact(self):
    self.assertEqual('n1-standard-16',
                     caip.cpu_ram_to_machine_type(16, 60 * xm.GiB))

  def test_cpu_ram_to_machine_type_highmem(self):
    self.assertEqual('n1-highmem-64',
                     caip.cpu_ram_to_machine_type(1, 415 * xm.GiB))

  def test_cpu_ram_to_machine_type_highcpu(self):
    self.assertEqual('n1-highcpu-64',
                     caip.cpu_ram_to_machine_type(63, 1 * xm.GiB))

  def test_cpu_ram_to_machine_type_too_high(self):
    with self.assertRaises(ValueError):
      caip.cpu_ram_to_machine_type(1000, 1000)


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from xmanager.xm import job_blocks


class JobBlocksTest(unittest.TestCase):

  def test_from_mapping(self):
    args = job_blocks.SequentialArgs.from_collection({
        'a': 1,
        'b': 2,
        'c': 3,
    })

    self.assertEqual(args.to_list(str), ['--a', '1', '--b', '2', '--c', '3'])

  def test_from_sequence(self):
    args = job_blocks.SequentialArgs.from_collection([1, 2, 3])

    self.assertEqual(args.to_list(str), ['1', '2', '3'])

  def test_from_none(self):
    args = job_blocks.SequentialArgs.from_collection(None)
    self.assertEqual(args.to_list(str), [])

  def test_merge(self):
    args = job_blocks.SequentialArgs.merge(
        map(job_blocks.SequentialArgs.from_collection, [[1], {
            'a': 'z',
            'b': 'y'
        }, [2], {
            'b': 'x',
            'c': 't'
        }, [3]]))

    self.assertEqual(
        args.to_list(str), ['1', '--a', 'z', '--b', 'x', '2', '--c', 't', '3'])

  def test_to_dict(self):
    args = job_blocks.SequentialArgs.merge(
        map(job_blocks.SequentialArgs.from_collection, [['--knob'], {
            1: False
        }]))

    self.assertEqual(args.to_dict(), {'--knob': True, '1': False})

  def test_to_list_bool(self):
    args = job_blocks.SequentialArgs.from_collection({'yes': True, 'no': False})

    self.assertEqual(args.to_list(str), ['--yes', '--nono'])


if __name__ == '__main__':
  unittest.main()
